package no.oslotuc.settings;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import se.sics.jasper.Jasper;

import java.io.*;
import java.lang.reflect.Field;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

public class SettingsLoader {


    public class SettingsWrapper {

        public String logDirectoryPath;
        public String busesTucPath;
        public String notFound404Path;
        public String javaLibraryPath;
        public int serverPort;
    }


    private static Logger logger = LogManager.getLogger(SettingsLoader.class.getName());
    public static final Path settingsPath = Paths.get("settings.json");

    public SettingsLoader() { }


    public void loadSettings() {
        GsonBuilder builder = new GsonBuilder();
        Gson gson = builder.create();



        if(!settingsPath.toFile().exists())
            generateSettingsFile();

        try(InputStreamReader reader = new InputStreamReader(new FileInputStream(settingsPath.toAbsolutePath().toString()), Charset.forName("utf-8"))) {

            SettingsWrapper settingsWrapper = gson.fromJson(reader, SettingsWrapper.class);
            loadStaticSettings(settingsWrapper);
            generateLogDirectory();

        } catch (FileNotFoundException e) {
            generateSettingsFile();
            logger.error(e);
            System.exit(1);
        } catch (IOException e) {
            logger.error(e);
            System.exit(1);
        }
    }


    private void generateLogDirectory() {
        File file = Paths.get(Settings.logDirectoryPath).toFile();

        if(!file.exists())
            file.mkdirs();
    }


    private void loadStaticSettings(SettingsWrapper settingsWrapper) {

        Settings.logDirectoryPath = settingsWrapper.logDirectoryPath;
        Settings.busesTuc4 = settingsWrapper.busesTucPath;
        Settings.javaLibraryPath = settingsWrapper.javaLibraryPath;
        Settings.notFoundPath = settingsWrapper.notFound404Path;
        Settings.serverPort = settingsWrapper.serverPort;

        if(Settings.javaLibraryPath == null || Settings.javaLibraryPath.isEmpty())
            return;

        try {
            System.setProperty("java.library.path", Settings.javaLibraryPath);

            // Hack to reset and reload the java.library.path property

            Field fieldSysPath;
            fieldSysPath = ClassLoader.class.getDeclaredField( "sys_paths" );
            fieldSysPath.setAccessible( true );
            fieldSysPath.set( null, null );

        } catch (Exception e) {
            logger.error(e);
        }

        try {
            Settings.prolog = Jasper.newProlog(null, null, Settings.busesTuc4);
        } catch (InterruptedException e) {
            logger.error("Could not load prolog server!", e);
        }
    }

    private void generateSettingsFile() {
        SettingsWrapper settingsWrapper = new SettingsWrapper();
        settingsWrapper.logDirectoryPath = "";
        settingsWrapper.busesTucPath = "";
        settingsWrapper.notFound404Path = "";
        settingsWrapper.javaLibraryPath = "";
        settingsWrapper.serverPort = 9001;

        try(OutputStreamWriter writer = new OutputStreamWriter(new FileOutputStream(settingsPath.toAbsolutePath().toString()), Charset.forName("utf-8"))) {
            GsonBuilder builder = new GsonBuilder();
            Gson gson = builder.setPrettyPrinting().create();
            String json = gson.toJson(settingsWrapper);

            writer.write(json);
            logger.info("Generated new settings file.");
        } catch (IOException e) {
            logger.error(e);
        }
    }
}
