package OsloTucClosestBusStopFinder;

import com.google.gson.GsonBuilder;
import com.google.gson.reflect.TypeToken;
import model.Address;
import model.HPL;
import org.apache.commons.io.IOUtils;

import java.io.*;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.Normalizer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class App
{


    private static final String hplPath = "C:\\Users\\Espen\\workspace\\togtuc\\src\\db\\tables\\r0320_141005\\reghpl.pl";
    private static final String statCoordPath = "C:\\Users\\Espen\\workspace\\togtuc\\src\\db\\statcoord2.pl";
    private static final String adressesPath = "C:\\Users\\Espen\\git\\OsloAdressCoordinatesFinder\\osloAdressesAndCoordinates.json";
    private static final String streetStatOutputPath = "regstr.pl";
    private static final String regCompStreetOutputPath = "regcompstr.pl";


    private static final Pattern hplPattern = Pattern.compile("^hpl\\('(\\d+)',(.*?),(.*?),'(.*?)'\\)\\.$");
    private static final Pattern statCoordPattern = Pattern.compile("^statcoord\\('(\\d+)',\\s+'(.*?)',\\s+(.*?),\\s+(.*?)\\)\\.$");

    private static Map<Integer, HPL> hpls = new HashMap<>();
    private static List<Address> addresses = new ArrayList<>();

    private static List<String> regCompStr = new ArrayList<>();


    public static void main( String[] args ) throws IOException {
        readHplFile();
        readStatCoordFile();
        readAddresses();

        System.out.println("Combining addresses with bus stops. Please wait...");

        combine();

        System.out.println("Operation finished!");
    }


    private static void readHplFile() throws IOException {

        List<String> lines = Files.readAllLines(Paths.get(hplPath), Charset.forName("utf-8"));

        for(String line : lines) {

            Matcher matcher = hplPattern.matcher(line);

            if(!matcher.find()) continue;

            int number = Integer.parseInt(matcher.group(1));
            String osloTucHplName = matcher.group(2);
            String name = matcher.group(3);

            hpls.put(number, new HPL(number, osloTucHplName, name));
        }
    }


    private static void readStatCoordFile() throws IOException {

        List<String> lines = Files.readAllLines(Paths.get(statCoordPath), Charset.forName("utf-8"));

        for(String line : lines) {

            Matcher matcher = statCoordPattern.matcher(line);

            if(!matcher.find()) continue;

            int number = Integer.parseInt(matcher.group(1).trim());
            String lat = matcher.group(3).trim();
            String lng = matcher.group(4).trim();

            if(!hpls.containsKey(number)) continue;

            HPL hpl = hpls.get(number);
            hpl.setLat(lat);
            hpl.setLng(lng);
        }
    }


    private static void readAddresses() throws IOException {
        String json = IOUtils.toString(new FileInputStream(adressesPath), Charset.forName("utf-8"));
        addresses = new GsonBuilder().create().fromJson(json, new TypeToken<List<Address>>(){}.getType());
    }


    private static void combine() throws IOException {


        StringBuilder regstrBuilder = new StringBuilder();

        regstrBuilder.append("/* -*- Mode:Prolog; coding:utf-8; -*- */\n");
        regstrBuilder.append("%% FILE regstr.pl\n");
        regstrBuilder.append("%% SYSTEM Busstuc\n");
        regstrBuilder.append("%% CREATED EJ-06102014 Automatic\n");
        regstrBuilder.append("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
        regstrBuilder.append(":-module(regstr, [\n");
        regstrBuilder.append("        streetstat/5\n");
        regstrBuilder.append("]).\n\n");

        StringBuilder regCompStrBuilder = new StringBuilder();
        regCompStrBuilder.append("/* -*- Mode:Prolog; coding:utf-8; -*- */\n");
        regCompStrBuilder.append("%% FILE regcompstr.pl\n");
        regCompStrBuilder.append("%% SYSTEM BussTUC\n");
        regCompStrBuilder.append("%% CREATED EJ-06102014 Automatic\n");
        regCompStrBuilder.append("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n");
        regCompStrBuilder.append(":-module(regcompstr, [\n");
        regCompStrBuilder.append("        composite_road/3\n");
        regCompStrBuilder.append("]).\n\n");



        for(Address address : addresses) {

            HPL closestHpl = getClosestHPL(address);

            String cleanCompressedStreetAddress = getCleanedAndCompressedAddress(address.name);
            String cleanStreetAddress = getCleanedAddress(address.name);
            int[] minMax = getMinMaxStreetNumber(address);

            String streetStat = String.format("streetstat(%s,'%s',%d,%d,%s).%s",
                    cleanCompressedStreetAddress,
                    cleanStreetAddress,
                    minMax[0],
                    minMax[1],
                    closestHpl.getOsloTucHplName(),
                    "\n");

            regstrBuilder.append(streetStat);
            regCompStrBuilder.append(createCompStr(cleanStreetAddress, cleanCompressedStreetAddress) + "\n");

        }

        FileOutputStream regStrWriter = new FileOutputStream(streetStatOutputPath);
        IOUtils.write(regstrBuilder.toString(), regStrWriter, Charset.forName("utf-8"));

        FileOutputStream regCompStrWriter = new FileOutputStream(regCompStreetOutputPath);
        IOUtils.write(regCompStrBuilder.toString(), regCompStrWriter, Charset.forName("utf-8"));
    }


    private static HPL getClosestHPL(Address address) {

        HPL closestHpl = null;
        double shortestDistance = Double.MAX_VALUE;

        for(HPL hpl : hpls.values()) {

            if(hpl.getLat() == null || hpl.getLat().isEmpty() || hpl.getLat().indexOf('.') == 1) continue;
            if(hpl.getLng() == null || hpl.getLng().isEmpty() || hpl.getLng().indexOf('.') == 1) continue;

            double distance = distanceFrom(
                    Float.parseFloat(address.lat),
                    Float.parseFloat(address.lng),
                    Float.parseFloat(hpl.getLat()),
                    Float.parseFloat(hpl.getLng())
            );

            if(distance < shortestDistance) {
                shortestDistance = distance;
                closestHpl = hpl;
            }
        }

        return closestHpl;
    }


    public static float distanceFrom(float lat1, float lng1, float lat2, float lng2) {

        double earthRadius = 6371; //kilometers

        double dLat = Math.toRadians(lat2-lat1);
        double dLng = Math.toRadians(lng2-lng1);

        double a =    Math.sin(dLat/2)
                    * Math.sin(dLat/2)
                    + Math.cos(Math.toRadians(lat1))
                    * Math.cos(Math.toRadians(lat2))
                    * Math.sin(dLng/2) * Math.sin(dLng/2);

        double c = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1-a));
        float dist = (float) (earthRadius * c);

        return dist;
    }


    private static String getCleanedAndCompressedAddress(String address) {

        //Save the åååååå
        String cleaned = address.replaceAll("å", "#");

        cleaned = getCleanedAddress(cleaned);
        cleaned = cleaned.replaceAll("#", "å");

        // Order counts, "veien" before "vei", etc
        cleaned = cleaned.replaceAll("veien", "_street");
        cleaned = cleaned.replaceAll("vei", "_street");
        cleaned = cleaned.replaceAll("vegen", "_street");
        cleaned = cleaned.replaceAll("veg", "_street");
        cleaned = cleaned.replaceAll("gate", "_street");
        cleaned = cleaned.replaceAll("gata", "_street");
        cleaned = cleaned.replaceAll("[\\.\\s\\-]", "_");
        cleaned = cleaned.replaceAll("^\\d+", "");
        cleaned = cleaned.replaceAll("^_+", "");
        cleaned = cleaned.replaceAll("_+", "_").toLowerCase();

        return cleaned;
    }

    private static String getCleanedAddress(String address) {

        //Save the åååååå
        address = address.replaceAll("[åÅ]", "#");
        address = deAccent(address).replaceAll("[\\.`´\\^¨~']", "");
        address = address.replaceAll("#", "å");
        return address;
    }


    private static int[] getMinMaxStreetNumber(Address address) {
        int[] minMax = new int[] { Integer.MAX_VALUE, Integer.MIN_VALUE };

        String numbersString = address.numbers.replaceAll("[A-Za-zæøåÆØÅ]", "").replaceAll("[-,]", " ").replaceAll("\\s+", " ");
        String[] numbers = numbersString.split("\\s+");

        for(String numberStr : numbers) {

            int number = Integer.parseInt(numberStr);

            if(number < minMax[0])
                minMax[0] = number;

            if(number > minMax[1])
                minMax[1] = number;
        }

        if(minMax[0] < 0) minMax[0] = 1;
        if(minMax[1] < 0) minMax[1] = 1;

        return minMax;
    }

    private static String deAccent(String str) {
        String nfdNormalizedString = Normalizer.normalize(str, Normalizer.Form.NFD);
        Pattern pattern = Pattern.compile("\\p{InCombiningDiacriticalMarks}+");
        return pattern.matcher(nfdNormalizedString).replaceAll("");
    }


    private static String createCompStr(String cleanAddress, String cleanCompressedStreetAddress) {

        cleanAddress = cleanAddress.replaceAll("^\\d+", "").replaceAll("\\s+", " ").replaceAll("-", " ").trim().toLowerCase();

        String[] addressSplit = cleanAddress.split("\\s+");

        String bracketedPart = "[";

        for(int i = 1; i < addressSplit.length; i++) {

            String singleSplit = addressSplit[i];

            switch (singleSplit) {
                case "veien":
                case "vei":
                case "vegen":
                case "veg":
                case "gate":
                case "gata":
                    singleSplit = "street";
            }

            bracketedPart += singleSplit;

            if(i != addressSplit.length -1)
                bracketedPart += ",";
        }

        bracketedPart += "]";

        return String.format("composite_road(%s,%s,%s).", addressSplit[0], bracketedPart, cleanCompressedStreetAddress);
    }
}
