package RegToppParser.control;

import RegToppParser.interfaces.ParseListener;
import RegToppParser.interfaces.RegToppFileParser;
import RegToppParser.models.FileType;
import RegToppParser.models.Version;
import RegToppParser.util.FileUtil;

import java.io.File;
import java.io.FileNotFoundException;
import java.nio.charset.Charset;
import java.nio.file.Path;
import java.util.*;

/**
 * Created on 15.11.14.
 * @author Espen Jacobsson
 */
public class ParseManager {

    private Version version;
    private Map<FileType, File> regToppFiles = new HashMap<>();
    private Path regToppFolder;

    private List<ParseListener> listeners = new ArrayList<>();



    public ParseManager(Path regToppFolder) {
        this.regToppFolder = regToppFolder;
    }


    public void registerListener(ParseListener listener) {
        listeners.add(listener);
    }


    public void registerListeners(ParseListener... listeners) {
        for(ParseListener listener : listeners) {
            this.listeners.add(listener);
        }
    }


    public void performParsing() {

        long now = System.currentTimeMillis();

        findFiles();
        findVersion();
        VersionManager.registerFileReaders(version);

        System.out.println("Version is ::: " + version);

        for(FileType fileType : regToppFiles.keySet()) {

            if(!regToppFiles.containsKey(fileType))
                continue;


            RegToppFileParser parser = FileTaskDelegator.getInstance().getFileParser(fileType);

            if(parser == null) {
                System.out.println("Could not find parser for file type: " + fileType);
                continue;
            }

            parser.parse(regToppFiles.get(fileType));
        }

        notifyAllFileParsersFinished();

        System.out.println("Parsing files took: " + (System.currentTimeMillis() - now) + "ms");

        for(ParseListener listener : listeners) {
            listener.ready(version);
        }
    }


    private void findFiles() {
        File[] files = FileFinder.findFiles(regToppFolder);

        for(File file : files) {
            String extension = FileUtil.getFileExtension(file);
            FileType fileType = null;

            try {
                fileType = FileType.valueOf(extension.toUpperCase());
            } catch(IllegalArgumentException iae) {
                System.out.println("Tried using a file type which isn't valid.");
            }

            if(fileType == null)
                continue;

            regToppFiles.put(fileType, file);
        }
    }


    private void findVersion() {

        if(!regToppFiles.containsKey(FileType.FRM)) {

            if(regToppFiles.containsKey(FileType.TMS)) {
                version = Version.V_1_2;
            }
            else if(regToppFiles.containsKey(FileType.TDA)) {
                version = Version.V_1_1_D;
            }

            return;
        }

        try {
            Scanner scanner = new Scanner(regToppFiles.get(FileType.FRM), Charset.forName("utf-8").toString());
            String line = scanner.nextLine().trim();

            if(line.startsWith("Ver=") || line.startsWith("ver=")) {
                String version = "V_" + line.substring(4).replaceAll("\\.", "_");
                this.version = Version.valueOf(version);
            }

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
    }


    private void notifyAllFileParsersFinished() {
        for(FileType fileType : regToppFiles.keySet()) {

            if(!regToppFiles.containsKey(fileType)) {
                continue;
            }

            RegToppFileParser parser = FileTaskDelegator.getInstance().getFileParser(fileType);

            if(parser == null) {
                System.out.println("Could not find parser for file type: " + fileType);
            }
        }
    }
}